@@ -49,57 +49,27 @@
 
 
 /* Various flags */
+/* See metadata-exported.h for the complete list. */
 /* Note that the bits no longer necessarily correspond to LVM1 disk format */
 
-//#define PARTIAL_VG		0x00000001U	/* VG */
-//#define EXPORTED_VG          	0x00000002U	/* VG PV */
-//#define RESIZEABLE_VG        	0x00000004U	/* VG */
-
 /* May any free extents on this PV be used or must they be left free? */
-//#define ALLOCATABLE_PV         	0x00000008U	/* PV */
 
-#define SPINDOWN_LV          	0x00000010U	/* LV */
-#define BADBLOCK_ON       	0x00000020U	/* LV */
-//#define VISIBLE_LV		0x00000040U	/* LV */
-//#define FIXED_MINOR		0x00000080U	/* LV */
-/* FIXME Remove when metadata restructuring is completed */
-//#define SNAPSHOT		0x00001000U	/* LV - internal use only */
-//#define PVMOVE			0x00002000U	/* VG LV SEG */
-//#define LOCKED			0x00004000U	/* LV */
-//#define MIRRORED		0x00008000U	/* LV - internal use only */
-#define VIRTUAL			0x00010000U	/* LV - internal use only */
-//#define MIRROR_LOG		0x00020000U	/* LV */
-//#define MIRROR_IMAGE		0x00040000U	/* LV */
-//#define MIRROR_NOTSYNCED	0x00080000U	/* LV */
-#define PRECOMMITTED		0x00200000U	/* VG - internal use only */
-//#define CONVERTING		0x00400000U	/* LV */
+#define SPINDOWN_LV          	UINT64_C(0x00000010)	/* LV */
+#define BADBLOCK_ON       	UINT64_C(0x00000020)	/* LV */
+#define VIRTUAL			UINT64_C(0x00010000)	/* LV - internal use only */
+#define PRECOMMITTED		UINT64_C(0x00200000)	/* VG - internal use only */
+#define POSTORDER_FLAG		UINT64_C(0x02000000) /* Not real flags, reserved for  */
+#define POSTORDER_OPEN_FLAG	UINT64_C(0x04000000) /* temporary use inside vg_read_internal. */
+#define VIRTUAL_ORIGIN		UINT64_C(0x08000000)	/* LV - internal use only */
 
-//#define MISSING_PV		0x00800000U	/* PV */
-//#define PARTIAL_LV		0x01000000U	/* LV - derived flag, not
-//						   written out in metadata*/
-
-#define POSTORDER_FLAG		0x02000000U /* Not real flags, reserved for  */
-#define POSTORDER_OPEN_FLAG	0x04000000U /* temporary use inside vg_read_internal. */
-#define VIRTUAL_ORIGIN		0x08000000U	/* LV - internal use only */
-
-//#define LVM_READ              	0x00000100U	/* LV VG */
-//#define LVM_WRITE             	0x00000200U	/* LV VG */
-//#define CLUSTERED         	0x00000400U	/* VG */
-#define SHARED            	0x00000800U	/* VG */
+#define SHARED            	UINT64_C(0x00000800)	/* VG */
 
 /* Format features flags */
-//#define FMT_SEGMENTS		0x00000001U	/* Arbitrary segment params? */
-//#define FMT_MDAS		0x00000002U	/* Proper metadata areas? */
-//#define FMT_TAGS		0x00000004U	/* Tagging? */
-//#define FMT_UNLIMITED_VOLS	0x00000008U	/* Unlimited PVs/LVs? */
-//#define FMT_RESTRICTED_LVIDS	0x00000010U	/* LVID <= 255 */
-//#define FMT_ORPHAN_ALLOCATABLE	0x00000020U	/* Orphan PV allocatable? */
 #define FMT_PRECOMMIT		0x00000040U	/* Supports pre-commit? */
-//#define FMT_RESIZE_PV		0x00000080U	/* Supports pvresize? */
-//#define FMT_UNLIMITED_STRIPESIZE 0x00000100U	/* Unlimited stripe size? */
 
-struct config_tree;
+struct dm_config_tree;
 struct metadata_area;
+struct alloc_handle;
 
 /* Per-format per-metadata area operations */
 struct metadata_area_ops {
@@ -402,6 +372,12 @@ struct pv_list *find_pv_in_pv_list(const struct dm_list *pl,
 /* Find LV segment containing given LE */
 struct lv_segment *find_seg_by_le(const struct logical_volume *lv, uint32_t le);
 
+/* Find pool LV segment given a thin pool data or metadata segment. */
+struct lv_segment *find_pool_seg(const struct lv_segment *seg);
+
+/* Find some unused device_id for thin pool LV segment. */
+uint32_t get_free_pool_device_id(struct lv_segment *thin_pool_seg);
+
 /*
  * Remove a dev_dir if present.
  */
@@ -445,6 +421,9 @@ int for_each_sub_lv(struct cmd_context *cmd, struct logical_volume *lv,
                     int (*fn)(struct cmd_context *cmd,
                               struct logical_volume *lv, void *data),
                     void *data);
+int move_lv_segments(struct logical_volume *lv_to,
+		     struct logical_volume *lv_from,
+		     uint64_t set_status, uint64_t reset_status);
 
 /*
  * Calculate readahead from underlying PV devices
@@ -454,10 +433,11 @@ void lv_calculate_readahead(const struct logical_volume *lv, uint32_t *read_ahea
 /*
  * For internal metadata caching.
  */
-int export_vg_to_buffer(struct volume_group *vg, char **buf);
+size_t export_vg_to_buffer(struct volume_group *vg, char **buf);
+int export_vg_to_config_tree(struct volume_group *vg, struct dm_config_tree **cft);
 struct volume_group *import_vg_from_buffer(const char *buf,
 					   struct format_instance *fid);
-struct volume_group *import_vg_from_config_tree(const struct config_tree *cft,
+struct volume_group *import_vg_from_config_tree(const struct dm_config_tree *cft,
 						struct format_instance *fid);
 
 /*
@@ -469,6 +449,24 @@ struct volume_group *import_vg_from_config_tree(const struct config_tree *cft,
  */
 int fixup_imported_mirrors(struct volume_group *vg);
 
+/*
+ * From thin_manip.c
+ */
+int attach_pool_metadata_lv(struct lv_segment *pool_seg,
+			    struct logical_volume *pool_metadata_lv);
+int attach_pool_data_lv(struct lv_segment *pool_seg,
+			struct logical_volume *pool_data_lv);
+int attach_pool_lv(struct lv_segment *seg, struct logical_volume *pool_lv,
+		   struct logical_volume *origin_lv);
+int detach_pool_lv(struct lv_segment *seg);
+int attach_pool_message(struct lv_segment *pool_seg, dm_thin_message_t type,
+			struct logical_volume *lv, uint32_t delete_id,
+			int auto_increment);
+int pool_has_message(const struct lv_segment *seg,
+		     const struct logical_volume *lv, uint32_t device_id);
+int extend_pool(struct logical_volume *lv, const struct segment_type *segtype,
+		struct alloc_handle *ah, uint32_t stripes, uint32_t stripe_size);
+
 /*
  * Begin skeleton for external LVM library
  */
