@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2001-2004 Sistina Software, Inc. All rights reserved.  
- * Copyright (C) 2004-2007 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2004-2011 Red Hat, Inc. All rights reserved.
  *
  * This file is part of the device-mapper userspace tools.
  *
@@ -15,7 +15,9 @@
 
 #include "dmlib.h"
 
-#ifdef DEBUG_MEM
+#ifdef VALGRIND_POOL
+#include "valgrind/memcheck.h"
+#endif
 
 #include <assert.h>
 #include <stdarg.h>
@@ -117,7 +119,9 @@ void *dm_malloc_aux_debug(size_t s, const char *file, int line)
 
 	/* log_debug("Allocated: %u %u %u", nb->id, _mem_stats.blocks_allocated,
 		  _mem_stats.bytes); */
-
+#ifdef VALGRIND_POOL
+	VALGRIND_MAKE_MEM_UNDEFINED(nb + 1, s);
+#endif
 	return nb + 1;
 }
 
@@ -143,11 +147,13 @@ void dm_free_aux(void *p)
 
 	/* sanity check */
 	assert(mb->magic == p);
-
+#ifdef VALGRIND_POOL
+	VALGRIND_MAKE_MEM_DEFINED(p, mb->length);
+#endif
 	/* check data at the far boundary */
-	ptr = ((char *) mb) + sizeof(struct memblock) + mb->length;
+	ptr = (char *) p + mb->length;
 	for (i = 0; i < sizeof(unsigned long); i++)
-		if (*ptr++ != (char) mb->id)
+		if (ptr[i] != (char) mb->id)
 			assert(!"Damage at far end of block");
 
 	/* have we freed this before ? */
@@ -167,9 +173,9 @@ void dm_free_aux(void *p)
 	mb->id = 0;
 
 	/* stomp a different pattern across the memory */
-	ptr = ((char *) mb) + sizeof(struct memblock);
+	ptr = p;
 	for (i = 0; i < mb->length; i++)
-		*ptr++ = i & 1 ? (char) 0xde : (char) 0xad;
+		ptr[i] = i & 1 ? (char) 0xde : (char) 0xad;
 
 	assert(_mem_stats.blocks_allocated);
 	_mem_stats.blocks_allocated--;
@@ -251,8 +257,6 @@ void dm_bounds_check_debug(void)
 	}
 }
 
-#endif
-
 void *dm_malloc_aux(size_t s, const char *file __attribute__((unused)),
 		    int line __attribute__((unused)))
 {
